/** @param {NS} ns */
export async function main(ns) {
	ns.clearLog()
	ns.tail()
	ns.disableLog('ALL')

	const MAKE_MONEY = "Human Trafficking"
	const VIGILANTE_JUSTICE = "Vigilante Justice"
	const TERRORISM = "Terrorism"
	const TRAIN_COMBAT = "Train Combat"
	const TERRITORY_WARFARE = "Territory Warfare"
	const maxMemberCount = 12
	const memberNamePool = [
		"CJ",
		"Big Smoke",
		"Sweet",
		"Ryder",
		"OG Loc",
		"Big Bear",
		"Emmet",
		"Big Devil",
		"Little Devil",
		"Tony",
		"LB",
		"Madd Dogg",
		"Little Weasel",
	]

	function estimateMemberPower(memberName) {
		const memberInfo = ns.gang.getMemberInformation(memberName)
		return Math.floor(Math.min(
			memberInfo.str_asc_mult,
			memberInfo.def_asc_mult,
			memberInfo.dex_asc_mult,
			memberInfo.agi_asc_mult,
		) / 1.5)
	}

	// Credit: Mysteyes. https://discord.com/channels/415207508303544321/415207923506216971/940379724214075442
	function getAscendTreshold(memberName) {
		const memberInfo = ns.gang.getMemberInformation(memberName)
		const mult = Math.min(
			memberInfo.str_asc_mult,
			memberInfo.def_asc_mult,
			memberInfo.dex_asc_mult,
			memberInfo.agi_asc_mult,
		)
		if (mult < 1.632) return 1.6326;
		if (mult < 2.336) return 1.4315;
		if (mult < 2.999) return 1.284;
		if (mult < 3.363) return 1.2125;
		if (mult < 4.253) return 1.1698;
		if (mult < 4.860) return 1.1428;
		if (mult < 5.455) return 1.1225;
		if (mult < 5.977) return 1.0957;
		if (mult < 6.496) return 1.0869;
		if (mult < 7.008) return 1.0789;
		if (mult < 7.519) return 1.073;
		if (mult < 8.025) return 1.0673;
		if (mult < 8.513) return 1.0631;
		return 1.0591;
	}

	function getMinAscensionResult(memberName) {
		const memberAscensionResult = ns.gang.getAscensionResult(memberName)
		return memberAscensionResult === undefined ? undefined : Math.min(
			memberAscensionResult.str,
			memberAscensionResult.def,
			memberAscensionResult.dex,
			memberAscensionResult.agi
		)
	}

	function isAscensionReady(memberName) {
		return getMinAscensionResult(memberName) >= getAscendTreshold(memberName)
	}

	function isRecruitementReady(relativePower, memberCount) {
		const res = relativePower / 3 > memberCount - 3 // TODO
		// ns.print(ascentionCount / 3)
		return res
	}

	function getGangWarWinChances() {
		const gangInfo = ns.gang.getGangInformation()
		const otherGangNames = Object.keys(ns.gang.getOtherGangInformation()).filter(x => gangInfo.faction != x && x.territory > 0)
		const chancesToWin = otherGangNames.map(x => ns.gang.getChanceToWinClash(x))
		return Math.min(...chancesToWin)
	}

	while (true) {
		const gangInfo = ns.gang.getGangInformation()

		ns.print("running loop")

		// check for and recruit new members if available
		while (ns.gang.canRecruitMember()) {
			const availableNames = memberNamePool.filter(x => memberNamePool.includes(x))
			const memberName = availableNames[Math.floor(Math.random() * availableNames.length)]
			ns.gang.recruitMember(memberName)
			ns.gang.setMemberTask(memberName, TRAIN_COMBAT)
		}
		
		const gangMembers = ns.gang.getMemberNames()

		// check for and ascend members who are ready
		for (const memberName of gangMembers) {
			if (isAscensionReady(memberName)) {
				// ns.print("ascending " + memberName)
				ns.gang.ascendMember(memberName)
			}
		}

		for (const memberName of gangMembers) {
			if (isAscensionReady(memberName)) {
				// ns.print("ascending " + memberName)
				ns.gang.ascendMember(memberName)
			}
		}

		// TODO: equipement


		// gang war
		if (getGangWarWinChances() >= 0.6) {
			// go to war
			ns.gang.setTerritoryWarfare(true)
		}
		else if (getGangWarWinChances() <= 0.55) {
			// disengage
			ns.gang.setTerritoryWarfare(false)
		}

		// score each member by their estimated power level
		let memberDatas = gangMembers.map(x => [x, estimateMemberPower(x)])
		memberDatas.sort((a, b) => b[1] - a[1])
		const targetRelativePower = memberDatas[0][1] - 1
		for (const memberData of memberDatas) {
			const memberName = memberData[0]
			const relativePower = memberData[1]

			ns.print(memberName)
			// ns.print(getMinAscensionResult(memberName) / getAscendTreshold(memberName))
			// ns.print(isAscensionReady(memberName))

			if (relativePower < targetRelativePower && relativePower < 40) {
				ns.print("assignment: catchup")
				// bring up newcomers to the same level than established members
				ns.gang.setMemberTask(memberName, TRAIN_COMBAT)
			}
			else if (getMinAscensionResult(memberName) / getAscendTreshold(memberName) < 0.9) { // TODO: arbitrary constant to be ajusted
				ns.print("assignment: level up after ascention")
				// train a member back up after ascention
				ns.gang.setMemberTask(memberName, TRAIN_COMBAT)
			}
			else if (gangInfo.wantedPenalty < 0) {
				ns.print("assignment: reduce wanted level")
				// reduce wanted level
				ns.print(gangInfo.wantedPenalty)
				ns.gang.setMemberTask(memberName, VIGILANTE_JUSTICE)
			}
			else if (gangMembers.length < maxMemberCount && isRecruitementReady(relativePower, gangMembers.length)) {
				ns.print("assignment: gain respect")
				// gain respect to recruit new member
				ns.gang.setMemberTask(memberName, TERRORISM)
			}
			else if (relativePower < 20) { // TODO
				ns.print("assignment: train to next ascension")
				// train up to next ascention
				ns.gang.setMemberTask(memberName, TRAIN_COMBAT)
			}
			else if (gangInfo.territory < 1 && getGangWarWinChances() < 0.95) {
				ns.print("assignment: gang war")
				// prepare for gang war
				ns.gang.setMemberTask(memberName, TERRITORY_WARFARE)
			}
			else {
				ns.print("assignment: make money")
				// go for the money
				ns.gang.setMemberTask(memberName, MAKE_MONEY) // TODO
			}
		}

		await ns.sleep(5000)
	}
}